from karamel.packages import *
from karamel.exception import KaramelException
from karamel.logger import logger

from .command import Command

__command__ = 'freeze'
__description__ = 'Freeze packages'

class FreezeCommand(Command):

    def __init__(self, packages_file_url, package_install_dir):
        super().__init__(__command__, __description__)
        self.add_argument('package', nargs='*')
        self.packages_file_url = packages_file_url
        self.package_install_dir = get_package_install_dir(package_install_dir)

    def callback(self, args):
        if args.package != []:
            package_freezed = freeze_packages(self.package_install_dir, args.package)
        else:
            package_freezed = freeze_packages(self.package_install_dir)

        for package, ref in package_freezed.items():
            print('{}=={}'.format(package, ref))



import urllib.request
import os, yaml, git, re

from karamel.exception import PackagesFileNotFound


empty = lambda *args, **kwargs: None


def read_packages_file(packages_file_url):
    try:
        response = urllib.request.urlopen(packages_file_url)
        content = response.read()
    except ValueError:
        try:
            stream = open(packages_file_url, 'r')
            content = stream.read()
            stream.close()
        except:
            raise PackagesFileNotFound(packages_file_url)

    return yaml.safe_load(content)


def get_package_install_dir(package_install_dir):
    package_install_dir = os.path.expanduser(package_install_dir)
    if package_install_dir != '' and not os.path.isdir(package_install_dir):
        os.makedirs(package_install_dir)
    return package_install_dir


def get_installed_packages(package_install_dir):
    return os.listdir(package_install_dir)


def search_packages(packages_file_url, pattern):
    packages = read_packages_file(packages_file_url)
    packages_name = list(packages.keys())
    r = re.compile('.*{}.*'.format(pattern))
    packages_matching_pattern = filter(r.match, packages_name)

    packages_found = {}
    for package_name in packages_matching_pattern:
        if package_name in packages:
            packages_found[package_name] = packages[package_name]

    return packages_found


def install_packages(packages_file_url,
                     package_install_dir,
                     packages_to_install,
                     on_package_downloading=empty,
                     on_package_installing=empty,
                     on_package_install_success=empty,
                     on_package_already_installed=empty,
                     on_package_not_found=empty,
                     on_package_bad_version_provided=empty):

    packages = read_packages_file(packages_file_url)
    installed_packages = get_installed_packages(package_install_dir)

    for package_name in packages_to_install:

        # TODO verify package_name with regex

        version_requested = None
        if '==' in package_name:
            package_name, version_requested = package_name.split('==')

        if package_name in packages:
            package = packages[package_name]
            package_path = os.path.join(package_install_dir, package_name)

            if not package_name in installed_packages:
                on_package_downloading(package_name)
                download_package(package_install_dir, package['url'])
                on_package_installing(package_name)
                on_package_install_success(package_name)
            else:
                on_package_already_installed(package_name, package_path)

            package_version = get_package_version(package_path)
            if version_requested and version_requested != package_version:
                try:
                    change_package_version(package_path, version_requested)
                except BadVersionProvided:
                    on_package_bad_version_provided(package_name, version_requested)

        else:
            on_package_not_found(package_name)


def freeze_packages(package_install_dir, packages_to_freeze=[]):
    packages_installed = get_installed_packages(package_install_dir)
    freeze = {}

    if packages_to_freeze == []:
        packages_to_freeze = packages_installed

    for package in packages_to_freeze:
        if package in packages_installed:
            package_path = os.path.join(package_install_dir, package)
            git_reference = get_package_version(package_path)
            freeze[package] = str(git_reference)
        else:
            freeze[package] = None

    return freeze


def download_package(package_install_dir, url):
    git.Git(package_install_dir).clone(url)


class BadVersionProvided(Exception):
    pass


def change_package_version(package_path, package_version):
    try:
        repo_git = git.Repo(package_path).git
        repo_git.checkout(package_version)
    except git.exc.GitCommandError as e:
        raise BadVersionProvided()


def get_package_version(package_path):
    repo = git.Repo(package_path)
    version = repo.head.commit

    current_tag = next((tag for tag in repo.tags if tag.commit == repo.head.commit), None)
    if current_tag: version = current_tag

    try: version = repo.head.ref
    except: pass

    return version



from karamel.packages import *
from karamel.exception import KaramelException
from karamel.logger import logger

from .command import Command

__command__ = 'list'
__description__ = 'List packages'

class ListCommand(Command):

    def __init__(self, packages_file_url, package_install_dir):
        super().__init__(__command__, __description__)
        self.add_argument('--installed', action='store_true')
        self.packages_file_url = packages_file_url
        self.package_install_dir = get_package_install_dir(package_install_dir)

    def callback(self, args):
        if args.installed:
            packages = get_installed_packages(self.package_install_dir)
        else:
            packages = read_packages_file(self.packages_file_url)

        for package_name in packages:
            logger.info('{}'.format(package_name))



from karamel.packages import *
from karamel.logger import logger

from .command import Command

__command__ = 'search'
__description__ = 'Search package'

class SearchCommand(Command):

    def __init__(self, packages_file_url):
        super().__init__(__command__, __description__)
        self.add_argument('pattern')
        self.packages_file_url = packages_file_url

    def callback(self, args):
        packages_found = search_packages(self.packages_file_url, args.pattern)

        for package_name, package_info in packages_found.items():
            logger.info('{}'.format(package_name))
            logger.info('  {}'.format(package_info['description']))
            logger.info('  {}'.format(package_info['url']))
            logger.info('')



from karamel.packages import *
from karamel.exception import KaramelException
from karamel.logger import logger

from .command import Command

__command__ = 'install'
__description__ = 'Install a package'

class InstallCommand(Command):

    def __init__(self, packages_file_url, package_install_dir):
        super().__init__(__command__, __description__)
        self.add_argument('package', nargs='*')
        self.add_argument('-f', '--freeze')
        self.packages_file_url = packages_file_url
        self.package_install_dir = get_package_install_dir(package_install_dir)

    def on_package_downloading(self, package_name):
        logger.info('Downloading \'{}\''.format(package_name))

    def on_package_installing(self, package_name):
        logger.info('Installing \'{}\''.format(package_name))

    def on_package_install_success(self, package_name):
        logger.info('Successfully installed \'{}\''.format(package_name))

    def on_package_already_installed(self, package_name, package_path):
        logger.info('Package \'{}\' already installed in \'{}\''.format(package_name, package_path))

    def on_package_not_found(self, package_name):
        logger.error('Package not found \'{}\''.format(package_name))

    def on_package_bad_version_provided(self, package_name, version):
        logger.error('Could not find the version \'{}\' for package \'{}\''.format(package_name, version))

    def callback(self, args):
        packages_to_install = args.package

        if args.freeze:
            with open(args.freeze, 'r') as stream:
                packages_freeze = [l.strip() for l in stream]
            packages_to_install += packages_freeze

        install_packages(
            self.packages_file_url,
            self.package_install_dir,
            packages_to_install,
            self.on_package_downloading,
            self.on_package_installing,
            self.on_package_install_success,
            self.on_package_already_installed,
            self.on_package_not_found,
            self.on_package_bad_version_provided)
